/*
 * topologyd.h - topologyd header file
 *
 * Copyright (C) 2020 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 */
#ifndef TOPOLOGYD_H
#define TOPOLOGYD_H

#include <time.h>
#include <libubox/uloop.h>

#include <easy/easy.h>
#include <timer_impl.h>
#include <cmdu.h>
#include <1905_tlvs.h>
#include <map_module.h>
#include <easymesh.h>

#include <cmdu_ackq.h>


#define IEEE1905_OBJECT		"ieee1905"
#define IEEE1905_EVENTS		"ieee1905"
#define TOPOLOGY_OBJECT		"topology"
#define I1905_OBJ 		"ieee1905"
#define map_plugin		"ieee1905.map"


#define TOPOLOGY_CONFIG		"topology"

#define TOPOLOGY_EVENT_ADD     "ieee1905.neighbor.add"
#define TOPOLOGY_EVENT_DEL     "ieee1905.neighbor.del"
#define MAP_AGENT_EVENTS       "ieee1905.cmdu"

#define TOPOLOGY_NODE_MAX	32
#define TOPOLOGY_NEIGHBOR_MAX	32
#define TOPOLOGY_VALIDITY	120   // 120 sec
#define TOPOLOGY_STATUS_INTERVAL 5
#define TOPOLOGY_MAX_ENTRY 32
#define TOPOLOGY_MAX_LINK 32
#define TOPOLOGY_MAX_BR_TUPLE 32
#define TOPOLOGY_NON_IEEE1905_NEIGHBOR_MAX 250

typedef uint32_t ieee1905_object_t;

#define OBJECT_INVALID	((uint32_t)-(1))

#define TOPOLOGY_DEPTH		16	/* max depth */
#define TOPOLOGY_REFRESH_INT	120	/* default refresh interval in secs */
#define TOPOLOGY_LOG_MAX	32	/* max entries in topology changelog */
#define L2_NBR_MAX_BEHIND_ADDR  20

#define MDNS_TXT_RECORDS_MAX 16
#define MDNS_NO_SERVICES_MAX 16
#define TOPO_NOTIFY_TIMER 4


enum ip6addr_type {
	IP6_TYPE_UNKNOWN,
	IP6_TYPE_LINKLOCAL,
	IP6_TYPE_DHCP,
	IP6_TYPE_STATIC,
	IP6_TYPE_SLAAC,
};

enum i1905_mediatype {
	I1905_802_3U_FAST_ETHERNET       = (0x0000),
	I1905_802_3AB_GIGABIT_ETHERNET   = (0x0001),
	I1905_802_11B_2_4_GHZ            = (0x0100),
	I1905_802_11G_2_4_GHZ            = (0x0101),
	I1905_802_11A_5_GHZ              = (0x0102),
	I1905_802_11N_2_4_GHZ            = (0x0103),
	I1905_802_11N_5_GHZ              = (0x0104),
	I1905_802_11AC_5_GHZ             = (0x0105),
	I1905_802_11AD_60_GHZ            = (0x0106),
	I1905_802_11AF_GHZ               = (0x0107),
	I1905_802_11AX_2_4_GHZ           = (0x0108),  /* 11ax */
	I1905_802_11AX_5_GHZ             = (0x0109),  /* 11ax */
	I1905_1901_WAVELET               = (0x0200),
	I1905_1901_FFT                   = (0x0201),
	I1905_MOCA_V1_1                  = (0x0300),
	I1905_MEDIA_UNKNOWN              = (0xFFFF),
};

struct text_record_t {
	char key[256];
	char value[256];
};

struct mdns_service_t {
	char instance_name[64];
	char application_protocol[16];
	char transport_protocol[16];
	char domain[256];
	char target[256];
	uint8_t ifindex;
	uint16_t port;
	uint8_t txt_no_entries;
	struct text_record_t *text_record[MDNS_TXT_RECORDS_MAX];
	time_t time;
	uint32_t ttl;
	uint16_t priority;
	uint16_t weight;
};

struct device_ipv4_t {
	uint8_t mac_address[6];
	uint8_t ipv4_addr[4];
	uint8_t ipv4_addr_type;
	uint8_t dhcp_addr[4];
	uint8_t mdns_no_services;
	struct mdns_service_t *mdns_service[16];
};

struct device_ipv6_t {
	uint8_t mac_address[6];
	uint8_t ipv6_addr[16];
	uint8_t ipv6_addr_type;
	uint8_t ipv6addr_origin[16];
	// TODO: ipv6 mdns support
	uint8_t mdns_no_services;
	struct mdns_service_t *mdns_service[16];
};

struct link_metric {
	uint8_t nbr_ifr_addr[6];
	uint8_t ieee802dot1br;
	uint32_t tx_packet_errors;
	uint32_t rx_packet_errors;
	uint32_t tx_packet_ok;
	uint32_t rx_packet_ok;
	uint16_t tx_max_xput;
	uint16_t tx_phy_rate;
	uint16_t tx_link_availability;
	uint8_t rx_rssi;
};

struct l2_nbr_t {
	uint8_t l2_intf_addr[6];
	uint8_t l2_nbr_addr[6];
	uint32_t behind_intf_nbr;
	uint8_t behind_intf_addr[L2_NBR_MAX_BEHIND_ADDR][6];
};

struct node_interface {
	uint8_t interface_id[6];
	uint16_t  media_type;
	uint8_t power_status;
	uint8_t generic_phy_oui[3];
	uint8_t generic_phy_variant;
	char generic_phy_url[1024];
	uint8_t media_specific_data_size;
	uint8_t network_membership[6];
	uint8_t role;
	uint8_t ap_channel_band;
	uint8_t freq_index1;
	uint8_t freq_index2;
};

struct nbr_data_t {
	uint8_t nbr_macid[6];
	uint32_t metric_num;
	struct link_metric ln_metric[TOPOLOGY_MAX_LINK];
};

struct bridge_tuple {
	uint32_t tuple_num;
	uint8_t tuple_macid[TOPOLOGY_MAX_BR_TUPLE][6];
};

struct node {
	uint8_t hwaddr[6];
	void *priv;
	struct hlist_node hlist;
	uint32_t nbr_num;
	time_t last_recv;
	char ingress_ifr_name[256];
	struct uloop_timeout validity_timer;
	uint32_t non1905_nbr_num;
	uint32_t l2_nbr_num;
	uint8_t non1905_nbr_localintf[6]; // local interface of the non1905 nbr
	uint8_t nbr_localintf[6]; // local interface of the 1905 nbr
	struct nbr_data_t nbr_list[TOPOLOGY_NEIGHBOR_MAX];
	uint8_t non1905_nbrlist[TOPOLOGY_NON_IEEE1905_NEIGHBOR_MAX][6];
	struct l2_nbr_t l2_nbr[TOPOLOGY_NEIGHBOR_MAX];
	//tr181 ieee1905device.
	uint8_t version;
	char friendly_name[128];
	char manufacturer_name[128];
	char manufacturer_model[128];
	char control_url[1024];
	char assoc_wifi_ref[1024];
	//IPv4 Address
	uint32_t ipv4_intf_nbr;
	struct device_ipv4_t device_ipv4[TOPOLOGY_MAX_ENTRY];
	//IPV6 Address
	uint32_t ipv6_intf_nbr;
	struct device_ipv6_t device_ipv6[TOPOLOGY_MAX_ENTRY];
	//Bridging tuple
	uint32_t bridge_tuple_num;
	struct bridge_tuple  br_tuple[TOPOLOGY_MAX_ENTRY];
	//local interfaces
	uint32_t local_intf_nbr;
	struct node_interface node_intf[TOPOLOGY_MAX_ENTRY];
};

struct topology_changelog {
	uint8_t rpt_macaddr[6];
	uint8_t rpt_ifmacaddr[6];
	enum { NBR_NODE_ADD, NBR_NODE_DEL } type;
	bool is1905_nbr;
	uint8_t nbr_macaddr[6];
	time_t timestamp;
};

#define NODE_HTABLE_SIZE	128
#define node_hash(n)	\
	((n[0] ^ n[1] ^ n[2] ^ n[3] ^ n[4] ^ n[5]) % NODE_HTABLE_SIZE)
#define CT_NODE_HTABLE_SIZE	128
#define ct_node_hash(n)		(n % 128)


struct topology {
	void *priv;
	uint32_t state;
	int32_t num_nodes;
	uint32_t num_changelog;
	struct topology_changelog *changelog;
	int32_t changelog_front;
	int32_t changelog_rear;
	struct hlist_head node_htable[NODE_HTABLE_SIZE];
};

struct topologyd_config {
	bool enabled;
	uint32_t depth;
	uint32_t refresh_int;
	uint32_t maxlog;
	uint8_t profile;
};

struct host_ntwk {
	void *priv;
	int32_t num_nodes;
	struct hlist_head node_htable[NODE_HTABLE_SIZE];
	struct hlist_head ct_node_htable[CT_NODE_HTABLE_SIZE];
};

struct topologyd_private {
	int debug;
	bool algo_running;
	enum { INCOMPLETE, AVAILABLE } status;
	struct ubus_context *ctx;
	struct uloop_fd uloop_fd;
	struct uloop_timeout heartbeat;
	struct uloop_timeout refresh_timer;
	struct uloop_timeout status_timer;
	struct ubus_object obj;
	struct topologyd_config config;
	ieee1905_object_t ieee1905;
	uint8_t ieee1905_macaddr[6];
	uint8_t albridge_macaddr[6];
	struct node selfnode;
	struct topology topo;
	struct host_ntwk host;
	struct uloop_timeout host_refresh_timer;
	struct uloop_timeout host_arp_timer;
	struct uloop_timeout host_wifi_timer;
	void *avahi_serv;
	void *simple_poll;
	struct json_object *ieee1905_interfaces;
	struct json_object *wifi_stations;
	struct json_object *network_dump;
	struct ubus_event_handler wifi_evh;
	bool is_mapagent;
	/* i1905 stack subscription */
	bool use_ieee1905;
	struct ubus_event_handler obj_register;
	uint32_t topo_oid;
	void *subscriber;
	bool subscribed;
	//Adding variable for mid matching
	struct cmdu_ackq cmdu_ack_q;
	int pending;
	//Adding global host variable
	struct host_config host_global_cfg;
	//Adding for topology notification refresh
	struct uloop_timeout topo_notify_timer;
};

struct topo_resp_msg {
	struct topologyd_private *priv;
	uint8_t origin[6];
	uint8_t source[6];
	uint16_t msg_mid;
	char ifname[16];
};


uint16_t send_cmdu(struct topologyd_private *priv, struct cmdu_buff *cmdu, uint8_t *dst);
void topologyd_algo_run(struct topologyd_private *p);
int topologyd_node_add(struct topologyd_private *priv, struct node *n);
int topologyd_get_ieee1905_macaddr(struct topologyd_private *p);
int send_buildcmdu(struct topologyd_private *priv, uint16_t msg_type);
struct cmdu_buff *topo_gen_topology_query(struct topologyd_private *priv, uint8_t profile);

//Function to send the query messages
int topologyd_send_ieee1905_topology_query(struct topologyd_private *p,
								uint8_t *dest);
int topologyd_send_ieee1905_query(struct topologyd_private *p, uint8_t *dest, uint16_t type, uint8_t *data);


void changelog_copy_node_info(struct topologyd_private *t, struct node *p,
				int32_t event, bool is1905);
void update_changelog_info(struct topologyd_private *priv, struct node *dest, struct node *src);
void topologyd_update_changelog_firstentry(struct topologyd_private *priv, struct node *dest);
void enqueue_changelog(struct topology *t, struct topology_changelog *c);
void update_ieee1905_node_info(struct topologyd_private *priv, struct node *dest, struct node *src);

//Functions to update the response message in node
void topologyd_process_topology_response(struct cmdu_buff *cstruct, struct topologyd_private *priv, char *itfr_name);
void topologyd_process_autoconfiguration_response(struct cmdu_buff *cstruct, struct topologyd_private *priv);
void topologyd_process_higherlayer_response(struct cmdu_buff *cstruct, struct topologyd_private *priv);
void topologyd_process_linkmetric_response(struct cmdu_buff *cstruct, struct topologyd_private *priv);
#if 0
void topologyd_process_generic_phy_response(struct cmdu_cstruct *cstruct, struct topologyd_private *priv);
#endif
void topology_add_generic_phy_data(uint8_t *phy, struct topologyd_private *priv);
int ieee1905_msg_from_ubus(struct blob_attr *msg, uint8_t *source, uint8_t *origin, uint16_t *msg_type, uint16_t *msg_mid, uint8_t **tlv, uint32_t *b_len, char *ingress_if_name);


//Function to get string from enum
const char *get_interface_type_string(uint16_t if_type);
const char *get_role_type_string(uint16_t role_type);
const char *get_registra_freq_string(uint32_t band);
const char *get_version(uint8_t ver);
const char *get_ipv4_type(uint8_t ipv4_type);
const char *get_ipv6_type(uint8_t ipv6_type);

//Function to send the node add or delete event
void topology_send_node_event(struct topology_changelog *c, char *ingress);

void topologyd_event_handler(void *c, struct blob_attr *msg);


bool config_init_host_data(char *package_name, char *section,
		struct topologyd_private *priv);

#endif /* TOPOLOGYD_H */
